/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>

#include "alias.h"
#include "skip-list.h"
#include "rli.h"
#include "ezconfig.h"
#include "dialog.h"

typedef struct alias_item_s {
    char *shortname;
    char *longname;
} alias_item_t;

struct alias_s {
    skiplist *list;

    char filename[MAXPATHLEN+1];

    time_t lastmodified;
};

static time_t lastmodified(const char *filename)
{
    struct stat st;
    int r;

    bzero ( &st, sizeof ( struct stat ) );
    r = stat(filename, &st);
    if (r) return 0;

    return st.st_mtime;
}

static int ch_compare(const void *a, const void *b)
{
    alias_item_t *aa = (alias_item_t *)a;
    alias_item_t *bb = (alias_item_t *)b;

    return strcmp(aa->shortname,bb->shortname);
}

static int parse_file(alias_t *al, const char *fname)
{
    FILE *stream;
    char line[1024];
    time_t last;

    /* JCS - made the newlists larger by default, since they were
     * overflowed as is.
     */
    /* we already have latest version */
    
    last=lastmodified(fname);

    /* If the file doesn't exist, lastmodified will be 0, so we make
       sure we have an empty skip list and return. */
    if (!last) {

      if (!al->list) {
	al->list = newlist(1000, 0.5, (int (*)(const void *, const void *))
			   &ch_compare);
      }

      return 0;
      
    }

    /* If the file exists but hasn't been modified since our last read. */
    if ((al->lastmodified >= last) && (al->lastmodified > 0)) {
	return 0;
    }

    /* slight race, results in another read later, doesn't matter */
    al->lastmodified = last;

    if (al->list) {
	destroylist(al->list);
    }

    al->list = newlist(1000, 0.5, (int (*)(const void *, const void *)) &ch_compare);
    
    stream = fopen(fname,"r");
    if (!stream) return 1;

    while (fgets(line, sizeof(line)-1,stream)) {
	char *s = line;
	alias_item_t *item;
	char *end;
	
	while((!isspace(*s)) && (*s != '\0')) s++;
	
	if (*s == '\0') continue;

	*s = '\0';
	s++;
	
	while (isspace(*s)) s++;

	end = s;
	while ((!isspace(*end)) && (*end != '\0') && (*end != '\n')) end++;
	*end = '\0';

	item = malloc(sizeof(alias_item_t));
	if (!item) return -1;

	item->shortname = xstrdup(line);
	item->longname = xstrdup(s);
	
	sinsert(al->list,(void *)item);
    }
        
    fclose(stream);

    return 0;
}

static char *get_filename(void)
{
    char *homedir = getenv("HOME");
    static char filename[MAXPATHLEN+1];
    
    if (!homedir) return ".aliases";

    snprintf(filename,sizeof(filename)-1,"%s/.aliases",homedir);

    return filename;
}
       
alias_t *alias_init(void)
{
    alias_t *ret;

    ret = malloc(sizeof(alias_t));
    if (!ret) return NULL;

    ret->lastmodified = 0;
    ret->list = NULL;

    parse_file(ret,get_filename());

    return ret;
}

char *alias_lookup(alias_t *al, char *name)
{
    alias_item_t *item;
    alias_item_t lookfor;

    if (al == NULL) {
	printf("alias_lookup() the alias_t is null");
	exit(0);
    }

    if (parse_file(al,al->filename)) {
	return NULL;
    }

    /* if there is no alias list just return */
    if (al->list == NULL) {
	return NULL;
    }
    
    lookfor.shortname = name;

    item = (alias_item_t *)ssearch(al->list, (void *)&lookfor);

    /* if found something return it */
    if (item) return item->longname;

    return NULL;
}

static int
exists(alias_t *al, char *shortname, char *longname)
{
    alias_item_t *item;
    skipnode *snode;

    for (item = (alias_item_t *)sfirst(al->list,&snode); item != NULL; 
	 item = (alias_item_t *)snext(&snode) ) {

	if (shortname) if (strcmp(item->shortname, shortname)) continue;
	if (longname) if (strcmp(item->longname, longname)) continue;

	return 1;
    }

    return 0;    
}



static void
add_item(alias_t *al, char *name, char *email)
{
    FILE *stream;
    alias_item_t *item;
 
    stream = fopen(get_filename(),"a+");

    if (!stream) {
	DLG_Alert(ALERT_warning,"Error opening alias file");
	return;
    }

    fprintf(stream,"%s\t\t\t%s\n",name,email);

    fclose(stream);

    item = malloc(sizeof(alias_item_t));
    if (!item) {
	DLG_Alert(ALERT_warning,"System error adding to alias file");	
	return;
    }
    
    item->shortname = xstrdup(name);
    item->longname = xstrdup(email);

    if ((!item->shortname) || (!item->longname)) {
	DLG_Alert(ALERT_warning,"System error adding to alias file");	
	return;
    }
	
    sinsert(al->list,(void *)item);
}

void alias_smart_add(alias_t *al, char *email)
{
    char result[256];
    char prompt[256];

    /* if smart aliasing not enabled don't do anything */
    if (config_getswitch("SmartAliasing",0) == 0) return;

    /* find if email already in there */
    if (exists(al,NULL,email)) return;

    /* ask to add it */


    result[0] = '\0';

    do {
	if (*result) {
	    snprintf(prompt,sizeof(prompt)-1,"Error: That alias name already exists!  Add alias to use for <[ %s ]> in the future",email);
	} else {
	    snprintf(prompt,sizeof(prompt)-1,"Add alias to use for <[ %s ]> in the future",email);
	}

	if (!DLG_GetStringFromUser(prompt,result)) return;
	if (!*result) return;

    } while (exists(al,result,NULL));
    
    add_item(al,result,email);
}
