/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * internal editing functions for EzMail
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>

#include "edit.h"
#include "common.h"
#include "comkeys.h"
#include "screen.h"
#include "text.h"
#include "menus.h"
#include "msgs.h"
#include "and_keys.h"
#include "windows.h"
#include "dialog.h"
#include "and_scr.h"
#include "ezconfig.h"
#include "clinet.h"

typedef struct key_table_R {
  int key;
  int always_useable;
  void (*keyfunc)();
} FKEYS, *FKEYS_pt;

extern int DoMenuCommand();

extern char *SCR_GetKeyText();

extern SCR_LINES_pt TXT_GetScrLine();
extern SCR_LINES_pt TXT_GetScrLineByRow();

int g_LastCmdIndex = -1;

void recalc_mark(TEXT_BUF_pt atbuf);

/*
 * Gap management routines.
 */

void EDT_InitGap(TEXT_BUF_pt atbuf)
{
  /* set the gap position and size, assume gap is at end of buffer */

  if (atbuf) {
    atbuf->gap_start = atbuf->fsize;
    atbuf->gap_size  = imax(0, atbuf->bsize - atbuf->fsize);
  }
}

long EDT_GetGapSize(TEXT_BUF_pt atbuf)
{
  return(atbuf->gap_size);
}

int EDT_SetGap(TEXT_BUF_pt atbuf,long char_pos)
{
  /* set the gap right after the char_pos. */

  if (atbuf == (TEXT_BUF_pt)NULL || atbuf->gap_size <= 0) return(-1);

  EDT_CloseGap(atbuf);

  if (atbuf->gap_start != char_pos) {
    int bytes = atbuf->fsize - char_pos;
    if (bytes > 0) {
      bcopy(&atbuf->buf[char_pos],
	    &atbuf->buf[char_pos + atbuf->gap_size], bytes);
      atbuf->gap_start = char_pos;
    }
  }

  return(0);
}

void EDT_GrowGap(TEXT_BUF_pt atbuf,int bytes)
{
  /* something got deleted from buffer.  If gap is at end of buffer,
     reinitialize the gap, else, assume the gap is growing downward,
     means that the GapStart position is not changed.
  */

  if (atbuf->gap_start < atbuf->fsize) atbuf->gap_size += bytes;
  else EDT_InitGap(atbuf);

  atbuf->modify_count++;
#ifndef IBMPC
  atbuf->ckpt_count++;
#endif
}

void EDT_CloseGap(TEXT_BUF_pt atbuf)
{
  /* move the gap to the end of buffer */

  if (TF_ISSET(atbuf->flags, TF_EDITING)) {
    if (atbuf->gap_size > 0 && atbuf->gap_start < atbuf->fsize) {
      int bytes = atbuf->fsize - atbuf->gap_start;
      if (bytes > 0) {
	bcopy(&atbuf->buf[atbuf->gap_start + atbuf->gap_size],
	      &atbuf->buf[atbuf->gap_start], bytes);
	EDT_InitGap(atbuf);
      }
    }
  }
}

void EDT_ShrinkGap(TEXT_BUF_pt atbuf,int bytes)
{
  /* someting was inserted into the gap, move the start and size */

  if (bytes > 0) {
    atbuf->gap_start += bytes;
    atbuf->gap_size  -= bytes;
    atbuf->modify_count++;
#ifndef IBMPC
    atbuf->ckpt_count++;
#endif
  }
}

int EDT_GetBytes(TEXT_BUF_pt atbuf,long pos,char *buf,int buflen)
{
  int save_buflen = buflen;

  if (atbuf->buf) {
    if (TF_ISSET(atbuf->flags, TF_EDITING)) {
      if (buflen > 0 && pos < atbuf->fsize) {
	int bytes;
	if (pos < atbuf->gap_start) {
	  bytes = imin(buflen, (atbuf->gap_start - pos));
	  bcopy(&atbuf->buf[pos], buf, bytes);
	  buflen -= bytes;
	  buf += bytes;
	  pos += bytes;
	}
	if (buflen > 0 && pos >= atbuf->gap_start) {
	  bytes = imin(buflen, imax(0, (atbuf->fsize - pos)));
	  bcopy(&atbuf->buf[atbuf->gap_size + pos], buf, bytes);
	  buflen -= bytes;
	}
      }
    }
    else {
      if (buflen > 0) {
	int bytes;
	bytes = imin(buflen, imax(0, (atbuf->fsize - pos)));
	bcopy(&atbuf->buf[pos - atbuf->offset], buf, bytes);
	buflen -= bytes;
      }
    }
  }

  return(save_buflen - buflen);
}

/*
 * The following routines perform specific key functions.
 */

long EDT_FindPrevPage(TEXT_BUF_pt atbuf)
{
  int nbytes, nlines;
  char *sptr, *eptr;
  WINFO *w = atbuf->awp->awin;

  eptr = &atbuf->buf[imax(0, atbuf->bop_pos - atbuf->offset - 1)];
  nbytes = imin(atbuf->bop_pos-atbuf->offset, w->data_height*w->data_width);
  sptr = &atbuf->buf[imax(0, (atbuf->bop_pos - atbuf->offset - nbytes))];

  nlines = w->data_height;
  nbytes = 0;
  if (*eptr == '\n') eptr--;

  while (nlines && (eptr > sptr)) {
    if (*eptr == '\n' || nbytes > w->data_width) {
      if (--nlines <= 0) {
	if (*eptr == '\n') eptr++;
	break;
      }
      nbytes = 0;
    }
    eptr--; nbytes++;
  }

  return((long)(atbuf->offset + (eptr - atbuf->buf)));
}

long EDT_find_next_eol(TEXT_BUF_pt atbuf,long pos)
{
  char buf[STR255LEN + 1];
  long new_pos = pos;

  while (new_pos < atbuf->fsize) {
    int i, bytes = EDT_GetBytes(atbuf, (long)new_pos, &buf[0], STR255LEN);
    if (bytes <= 0) break;
    buf[bytes] = 0;
    for (i=0; i<bytes; i++)
      if (buf[i] == '\n') {
	bytes = i;
	break;
      }
    if (buf[bytes] == '\n') {
      new_pos += bytes;
      break;
    }
    new_pos = imin(atbuf->fsize, (new_pos + STR255LEN));
  }
  return(new_pos);
}

long EDT_find_last_eol(TEXT_BUF_pt atbuf,long pos)
{
  char buf[STR255LEN + 1];
  long new_pos = pos;

  while (new_pos > 0) {
    int bytes;
    new_pos = imax(0, (new_pos - STR255LEN));
    bytes = EDT_GetBytes(atbuf, (long)new_pos, &buf[0], (int)(pos - new_pos));
    if (bytes <= 0) break;
    buf[bytes] = 0;
    pos -= bytes;
    while (--bytes > 0 && buf[bytes] != '\n');
    if (buf[bytes] == '\n') {
      new_pos += (bytes + 1);
      break;
    }
  }
  return(new_pos);
}

void EDT_EndCursor(TEXT_BUF_pt atbuf)
{
  atbuf->char_pos = EDT_find_next_eol(atbuf, (long)atbuf->char_pos);
  recalc_mark(atbuf);
}

void EDT_HomeCursor(TEXT_BUF_pt atbuf)
{
  atbuf->char_pos = EDT_find_last_eol(atbuf, (long)atbuf->char_pos);

  if (atbuf->char_pos < atbuf->hdr_end) {
    char *colon_pos;
    long  pos = atbuf->char_pos;
    EDT_CloseGap(atbuf);
    colon_pos = index(&atbuf->buf[pos], ':');
    if (colon_pos) {
      char *next_line = index(&atbuf->buf[pos], '\n');
      if (next_line && next_line > colon_pos) {
	pos = colon_pos - &atbuf->buf[0] + 1;
	if (atbuf->buf[pos] == SPC_KEY || atbuf->buf[pos] == TAB_KEY)
	  pos++;
      }
      atbuf->char_pos = pos;
    } else EDT_EndCursor(atbuf);
  }

  recalc_mark(atbuf);
}

void EDT_CursorRight(TEXT_BUF_pt atbuf)
{
  if (atbuf->char_pos < atbuf->fsize) {
    atbuf->char_pos++;
    recalc_mark(atbuf);
  } else {
    SCR_ClearKeyboard();
    SCR_Beep();
  }
}

void EDT_CursorLeft(TEXT_BUF_pt atbuf)
{
  if (atbuf->char_pos > 0) {
    atbuf->char_pos--;
    recalc_mark(atbuf);
  } else {
    SCR_ClearKeyboard();
    SCR_Beep();
  }
}

void EDT_CursorUp(TEXT_BUF_pt atbuf)
{
  SCR_LINES_pt cur_line = TXT_GetScrLine(atbuf, (long)atbuf->char_pos);

  if (cur_line) {
    long new_pos;
    cur_line = TXT_GetScrLineByRow(atbuf, (int)(cur_line->row - 1));
    if (cur_line)
      new_pos = cur_line->char_pos +
	imax(0, imin(cur_line->len-1, atbuf->x_cursor));
    else
      new_pos = imax(0, atbuf->bop_pos - 1);

    atbuf->char_pos = new_pos;
    if (new_pos < atbuf->hdr_end) EDT_EndCursor(atbuf);
    else if (cur_line == (SCR_LINES_pt)NULL) EDT_HomeCursor(atbuf);
    recalc_mark(atbuf);
  }
}

void EDT_CursorDown(TEXT_BUF_pt atbuf)
{
  SCR_LINES_pt cur_line = TXT_GetScrLine(atbuf, (long)atbuf->char_pos);

  if (cur_line) {
    long new_pos = cur_line->char_pos + cur_line->len;
    cur_line = TXT_GetScrLineByRow(atbuf, cur_line->row + 1);
    if (cur_line)
      new_pos = cur_line->char_pos +
	imax(0, imin((cur_line->len - 1), atbuf->x_cursor));

    atbuf->char_pos = imin(new_pos, atbuf->fsize);
    if (new_pos < atbuf->hdr_end) EDT_HomeCursor(atbuf);
    recalc_mark(atbuf);
  }
}

void EDT_PageUp(TEXT_BUF_pt atbuf)
{
  if (TF_ISSET(atbuf->flags, TF_OFFDSK) && atbuf->offset > 0) {
    int nbytes = atbuf->bop_pos - atbuf->offset;
    if (nbytes < screen_cols * screen_rows) {
      nbytes = atbuf->bsize - nbytes;
      nbytes = TXT_PageFile(atbuf, (long)(imax(0, (atbuf->offset - nbytes))));
      atbuf->bop_pos = atbuf->offset + nbytes;
    }
  }

  if (atbuf->char_pos <= 0 || atbuf->bop_pos <= 0) {
    SCR_ClearKeyboard();
    SCR_Beep();
  } else {
    EDT_CloseGap(atbuf);
    atbuf->char_pos = EDT_FindPrevPage(atbuf);
  }
  recalc_mark(atbuf);
}

void EDT_PageDown(TEXT_BUF_pt atbuf)
{
  if (atbuf->eop_pos >= atbuf->fsize) {
    SCR_ClearKeyboard();
    SCR_Beep();
  } else
    atbuf->char_pos = atbuf->eop_pos;

  recalc_mark(atbuf);
}

void EDT_Home(TEXT_BUF_pt atbuf)
{
  if (TF_ISSET(atbuf->flags, TF_OFFDSK) && atbuf->offset) {
    atbuf->bop_pos = atbuf->offset + imin(atbuf->bsize, atbuf->fsize);
  }

  if (atbuf->char_pos <= 0) {
    SCR_ClearKeyboard();
    SCR_Beep();
  } else
    atbuf->char_pos = 0;

  recalc_mark(atbuf);
}

void EDT_End(TEXT_BUF_pt atbuf)
{
  if (TF_ISSET(atbuf->flags, TF_OFFDSK)) {
    if ((atbuf->offset + atbuf->bsize) <= (atbuf->fsize - 1))
      TXT_PageFile(atbuf, (long)(atbuf->fsize - 1));
  }

  if (atbuf->char_pos >= atbuf->fsize && atbuf->eop_pos >= atbuf->fsize) {
    SCR_ClearKeyboard();
    SCR_Beep();
  } else 
    atbuf->char_pos = atbuf->fsize;

  recalc_mark(atbuf);
}

void EDT_DeleteChar(TEXT_BUF_pt atbuf)
{
  if (atbuf->char_pos < atbuf->fsize) {
    EDT_CloseGap(atbuf);
    bcopy(&atbuf->buf[atbuf->char_pos + 1],
	  &atbuf->buf[atbuf->char_pos],
	  (int)(atbuf->fsize - atbuf->char_pos));
    atbuf->fsize--;
    EDT_GrowGap(atbuf, 1);
    if (atbuf->char_pos <= atbuf->hdr_end) EDT_RecalcHeaders(atbuf);
    TXT_UpdateLine(atbuf, (long)atbuf->char_pos,
		   (int)atbuf->awp->awin->data_width, TRUE);
  } else {
    SCR_ClearKeyboard();
    SCR_Beep();
  }
}

void EDT_BackSpace(TEXT_BUF_pt atbuf)
{
  if (atbuf->char_pos > 0 && atbuf->char_pos <= atbuf->fsize) {
    atbuf->char_pos--;
    EDT_DeleteChar(atbuf);
  } else {
    SCR_ClearKeyboard();
    SCR_Beep();
  }
}

void EDT_KillLine(TEXT_BUF_pt atbuf)
{
  if (atbuf->char_pos < atbuf->fsize) {
    int bytes;
    int do_append = FALSE;
    int enable_paste = TRUE;
    EDT_CloseGap(atbuf);
    if (atbuf->mark_start >= 0 && atbuf->mark_end > atbuf->mark_start) {
      bytes = imax(0, atbuf->mark_end - atbuf->mark_start);
      atbuf->char_pos = atbuf->mark_start;
      atbuf->mark_start = atbuf->mark_end = atbuf->lastmark = -1;
      recalc_mark(atbuf);
    } else {
      if (EDT_KeyIndex(KEYCODE_CTRL_K) == g_LastCmdIndex) do_append = TRUE;
      bytes = EDT_find_next_eol(atbuf, (long)atbuf->char_pos)-atbuf->char_pos;
    }
    if (bytes > 0) {
      if (EDT_CopyText(&atbuf->buf[atbuf->char_pos],
		       bytes, do_append)  != 0) return;
      bcopy(&atbuf->buf[atbuf->char_pos + bytes],
	    &atbuf->buf[atbuf->char_pos],
	    (int)(atbuf->fsize - atbuf->char_pos - bytes));
      atbuf->fsize -= bytes;
      EDT_GrowGap(atbuf, bytes);
      if (atbuf->char_pos <= atbuf->hdr_end) EDT_RecalcHeaders(atbuf);
      TXT_UpdateLine(atbuf, (long)atbuf->char_pos,
		     (int)atbuf->awp->awin->data_width, TRUE);
    } else if (bytes == 0 && atbuf->buf[atbuf->char_pos] == '\n') {
      if (EDT_CopyText(&atbuf->buf[atbuf->char_pos],1,do_append) == 0)
	EDT_DeleteChar(atbuf);
    } else {
      SCR_ClearKeyboard();
      SCR_Beep();
      enable_paste = FALSE;
    }
    if (enable_paste) MNU_EnableItem(M_PASTE, TRUE);
  } else {
    SCR_ClearKeyboard();
    SCR_Beep();
  }
}

void recalc_mark(TEXT_BUF_pt atbuf)
{
  if (atbuf->mark_start >= 0) {
    atbuf->lastmark = atbuf->mark_end;
    atbuf->mark_end = atbuf->char_pos;
    if (atbuf->mark_start > atbuf->mark_end) {
      long swap_it = atbuf->mark_start;
      atbuf->mark_start = atbuf->mark_end;
      atbuf->mark_end = swap_it;
    }
  }
  MNU_EnableItem(M_COPY, (atbuf->mark_end > atbuf->mark_start ? TRUE : FALSE));
}

void EDT_SetMark(TEXT_BUF_pt atbuf)
{
  atbuf->mark_start = atbuf->char_pos;
  atbuf->mark_end = atbuf->mark_start;
  SCR_WriteStatusLine(MSG_MarkSet);
}

void EDT_ReturnKey(TEXT_BUF_pt atbuf)
{
  if (atbuf->char_pos < atbuf->hdr_end) {
    int len = atbuf->char_pos;
    EDT_CloseGap(atbuf);
    while(--len >= 0)
      if (atbuf->buf[len] != SPC_KEY || atbuf->buf[len] != TAB_KEY)
	break;
    if (len >= 0 && atbuf->buf[len] == ',')
      EDT_InsertString(atbuf, "\n  ", 3, TRUE);
    else {
      EDT_CursorDown(atbuf);
      if (atbuf->buf[atbuf->char_pos] != '\n') EDT_EndCursor(atbuf);
      else if (atbuf->buf[atbuf->char_pos - 1] == '\n')
	EDT_CursorDown(atbuf);
    }
  }
  else EDT_InsertChar(atbuf, (int)'\n', TRUE);
}

void EDT_RestoreLine(TEXT_BUF_pt atbuf)
{
  int   len;
  char *bufptr;

  if (EDT_GetCopyText(&bufptr, &len) == 0) {
    if (len >= (atbuf->bsize - atbuf->fsize)) {
      if (DLG_GetBooleanFromUser("Copy Buffer too large.  Paste anyway?",TRUE))
	len = imax(0, atbuf->bsize - atbuf->fsize);
      else len = 0;
    }
    if (len > 0) EDT_InsertString(atbuf, bufptr, len, FALSE);
  } else {
    SCR_ClearKeyboard();
    SCR_Beep();
  }
}

void EDT_TabKey(TEXT_BUF_pt atbuf)
{
#define TABLEN 4

  static char tab_str[TABLEN] = {SPC_KEY, SPC_KEY, SPC_KEY, SPC_KEY};

  if (atbuf->char_pos < atbuf->hdr_end) {
    EDT_CursorDown(atbuf);
    if (atbuf->char_pos == atbuf->hdr_end) EDT_CursorDown(atbuf);
  } else
    EDT_InsertString(atbuf, tab_str, TABLEN, TRUE);
}

void EDT_Copy(TEXT_BUF_pt atbuf)
{
  long start = imin(atbuf->mark_start, atbuf->fsize);
  long end   = imin(atbuf->mark_end, atbuf->fsize);

  if (start < end) {
    int bytes;
    EDT_CloseGap(atbuf);
    start = imax(0, start - atbuf->offset);
    end   = imax(0, end - atbuf->offset);
    bytes = imin(end - start, atbuf->fsize - atbuf->offset);
    if (EDT_CopyText(&atbuf->buf[start], bytes, FALSE))
      DLG_Alert(1, MSG_NotEnoughMemory);
    else SCR_WriteStatusLine(MSG_CopyDone);
    atbuf->mark_start = atbuf->mark_end = atbuf->lastmark = -1;
  } else {
    SCR_ClearKeyboard();
    SCR_Beep();
  }
}

void EDT_GotoMenu(TEXT_BUF_pt atbuf)
{
  int key;
  EDT_CloseGap(atbuf);
  SCR_WriteStatusLine(MSG_ShiftTabMsg);
  key = SCR_GetKey();
  MNU_MenuMode(key);
  if (gNovice) TXT_Hints(atbuf);
  else SCR_ClearStatusLine();
}

FKEYS g_EditKeys[] = {
  {KEYCODE_FAKERETURN, FALSE, EDT_ReturnKey},
  {KEYCODE_RETURN,     FALSE, EDT_ReturnKey},
  {KEYCODE_TAB,        FALSE, EDT_TabKey},
  {KEYCODE_BACKSPACE,  FALSE, EDT_BackSpace},
  {KEYCODE_DEL,        FALSE, EDT_BackSpace},
  {KEYCODE_DELETE,     FALSE, EDT_DeleteChar},
  {KEYCODE_UP,         TRUE,  EDT_CursorUp},
  {KEYCODE_DOWN,       TRUE,  EDT_CursorDown},
  {KEYCODE_RIGHT,      TRUE,  EDT_CursorRight},
  {KEYCODE_LEFT,       TRUE,  EDT_CursorLeft},
  /* {KEYCODE_HOME,       TRUE,  EDT_HomeCursor},
     {KEYCODE_END,        TRUE,  EDT_EndCursor}, */
  {KEYCODE_HOME,       TRUE,  EDT_GotoBOL},
  {KEYCODE_END,        TRUE,  EDT_GotoEOL},
  {KEYCODE_PAGE_UP,    TRUE,  EDT_PageUp},
  {KEYCODE_PAGE_DOWN,  TRUE,  EDT_PageDown},
  {KEYCODE_ESCAPE_HOME,TRUE,  EDT_Home},
  {KEYCODE_ESCAPE_END, TRUE,  EDT_End},
  {KEYCODE_CTRL_K,     FALSE, EDT_KillLine},
  {KEYCODE_CTRL_Y,     FALSE, EDT_RestoreLine},
  {KEYCODE_SET_MARK,   TRUE,  EDT_SetMark},
  {KEYCODE_COPY,       TRUE,  EDT_Copy},
  {KEYCODE_SHIFT_TAB,  FALSE, EDT_GotoMenu},
  {KEYCODE_CTRL_S,     TRUE,  EDT_SaveDraft}
};

#define g_EditKeysSize (sizeof(g_EditKeys) / sizeof(FKEYS))

int EDT_KeyIndex(int akey)
{
  int i;
  int upper_key = SCR_UpCaseRepresentation(akey);

  for (i=0; i<g_EditKeysSize; i++)
    if (g_EditKeys[i].key == akey || g_EditKeys[i].key == upper_key) return(i);

  return(-1);
}

int EDT_DoMenuCommand(int menuid)
{
  extern TEXT_VIEW_pt TXT_GetTLForFrontWindow();
  TEXT_VIEW_pt tview = TXT_GetTLForFrontWindow();

  if (tview) {
    TEXT_BUF_pt atbuf = tview->a_tbuf;
    EDT_CloseGap(atbuf);
    if (atbuf->modify_count > 0) {
      TF_SET_CHANGED(atbuf->flags);
      atbuf->modify_count = 0;
    }
    if (TF_ISSET(atbuf->flags, TF_EDITING)) {
      char astr[STR255LEN];
      astr[0] = 0;
      switch(menuid) {
      case M_ADD_HEADERS:
	atbuf->char_pos = 0;
	EDT_InsertString(atbuf,compose_headers,strlen(compose_headers),FALSE);
	atbuf->char_pos = 0;
	EDT_EndCursor(atbuf);
	break;
      case M_ADD_REPLY_TO:
	strcpy(astr, do_getuserid(NULL));
	if (get_local_string(MSG_REPLY_ADDRESS, astr))
	  EDT_AddHeader(atbuf, HDR_Reply_To, astr);
	break;
      case M_ADD_ACK:
	strcpy(astr, do_getuserid(NULL));
	if (get_local_string(MSG_ACK_ADDRESS, astr))
	  EDT_AddHeader(atbuf, HDR_Ack_to, astr);
	break;
      case M_ADD_SUB_INVITE:
	if (get_local_string(MSG_SUB_ADDRESS, astr))
	  EDT_AddHeader(atbuf, HDR_DirectoryCreation, astr);
	break;
      case M_ADD_DIST_INVITE:
	if (get_local_string(MSG_DIST_ADDRESS, astr))
	  EDT_AddHeader(atbuf, HDR_Redistribution_To, astr);
	break;
      case M_ADD_FAXHDRS:
	{
	  int use_more_options;
	  if (DLG_FaxHeaders(&use_more_options))
	    EDT_AddFaxHeaders(atbuf, use_more_options);
	}
	break;
      case M_CUT:
	EDT_KillLine(atbuf);
	break;
      case M_PASTE:
	EDT_RestoreLine(atbuf);
	break;
      default:
	goto not_editing_menu_command;
      }
      TXT_Redisplay(atbuf, atbuf->scr_lines);
      TXT_SetCursor(atbuf);
      return(TRUE);
    }
  }
not_editing_menu_command:
  return(DoMenuCommand(menuid));
}

void EDT_SetCursorInBody(TEXT_BUF_pt atbuf)
{
  if (atbuf) {
    if (atbuf->hdr_end > 0)
      atbuf->char_pos = imin(atbuf->hdr_end + 1, atbuf->fsize);
    else {
      atbuf->char_pos = 0;
      EDT_HomeCursor(atbuf);
    }
    TXT_SetCursor(atbuf);
  }
}

/* JCS - Set the cursor at the end of
 *   To: <--- 
 * because this makes it much easier to deal with forwarding
 * messages because you want to write the recipient's email 
 * address first!
 */
void EDT_SetCursorInHeader(TEXT_BUF_pt atbuf)
{
    if (atbuf) { 
	if (atbuf->hdr_end > 4) {
	    atbuf->char_pos = 4;
	} else {
	    atbuf->char_pos = 0;
	    EDT_HomeCursor(atbuf);
	}
	TXT_SetCursor(atbuf);
    }
}

int EDT_Edit(TEXT_BUF_pt atbuf)
{
  if (atbuf->fsize > atbuf->bsize) {
    DLG_Alert(1, MSG_FileTooBigToEdit);
    return(-1);
  }

  EDT_InitGap(atbuf);

  WIN_LockCursor(TRUE);
  TF_SET_EDITING(atbuf->flags);
  atbuf->modify_count = 0;
  {
    char ban_buf[STR255LEN];
    sprintf(ban_buf, "Editing -- Press %s to Exit Editor, ",
	    SCR_GetKeyText(KEYCODE_ESCAPE));
    strcat(&ban_buf[0], SCR_GetKeyText(KEYCODE_SHIFT_TAB));
    strcat(&ban_buf[0], " to Access Menu Bar.");
    TXT_SetBanner(atbuf, &ban_buf[0]);
  }
  EDT_RecalcHeaders(atbuf);

  if (atbuf->char_pos < atbuf->bop_pos || atbuf->char_pos >= atbuf->eop_pos)
    atbuf->char_pos = atbuf->bop_pos;

  if (atbuf->char_pos < atbuf->hdr_end) EDT_HomeCursor(atbuf);

  MNU_SetMenuCmdProc(EDT_DoMenuCommand);

  WIN_DoSetupMenus();
  TXT_SetCursor(atbuf);

  return 0;
}

void EDT_EditEnd(TEXT_BUF_pt atbuf)
{
  EDT_CloseGap(atbuf);
  TF_CLEAR(atbuf->flags, TF_EDITING);
  WIN_LockCursor(FALSE);
  TXT_SetBanner(atbuf, doc_standard_banner);
  if (atbuf->modify_count) {
    TF_SET_CHANGED(atbuf->flags);
    atbuf->modify_count = 0;
  }
  MNU_SetMenuCmdProc(DoMenuCommand);
  WIN_DoSetupMenus();
  WIN_HomeCursor();
}

int EDT_ProcessKey(int key,TEXT_BUF *atbuf)
{
  if (TF_ISSET(atbuf->flags, TF_EDITING)) {
    if (key == KEYCODE_ESCAPE)
      EDT_EditEnd(atbuf);
    else {
      int index = EDT_KeyIndex(key);
      if (index >= 0) {
	(*g_EditKeys[index].keyfunc)(atbuf);
	g_LastCmdIndex = index;
	TXT_SetCursor(atbuf);
      } else {
	  if (isascii(key) && isprint(key)) {
	      TF_SET_CHANGED(atbuf->flags);
	      EDT_InsertChar(atbuf, key, TRUE);
	  } else
	  return(FALSE);
      }
    }
  } else {
    int index = EDT_KeyIndex(key);
    if (index >= 0 && g_EditKeys[index].always_useable) {
      (*g_EditKeys[index].keyfunc)(atbuf);
      g_LastCmdIndex = index;
      TXT_SetCursor(atbuf);
    } else
      return(FALSE);
  }
  return(TRUE);
}

void EDT_InsertChar(TEXT_BUF_pt atbuf,int key,int set_cursor)
{
  long char_pos = atbuf->char_pos;
  if (EDT_SetGap(atbuf, (long)char_pos) == 0) {
    atbuf->buf[atbuf->char_pos++] = key;
    atbuf->fsize++;
    EDT_ShrinkGap(atbuf, 1);
    TXT_UpdateLine(atbuf, (long)char_pos, 1, TRUE);
    if (set_cursor) TXT_SetCursor(atbuf);
    if (char_pos <= atbuf->hdr_end) EDT_RecalcHeaders(atbuf);
  } else {
    /* Clearing the keyboard buffer stops a continuous stream of 
     * beeps and reduces damage that a spurrious paste can do 
     */
    SCR_ClearKeyboard();
    SCR_Beep();
  }
}

void EDT_InsertString(TEXT_BUF_pt atbuf,char *str,int str_len, int set_cursor)
{
  long char_pos = atbuf->char_pos;
  if ((EDT_SetGap(atbuf, (long)char_pos)==0) &&
      (EDT_GetGapSize(atbuf) >= str_len)) {
    bcopy(str, &atbuf->buf[char_pos], str_len);
    atbuf->char_pos += str_len;
    atbuf->fsize += str_len;
    EDT_ShrinkGap(atbuf, str_len);
    TXT_UpdateLine(atbuf, (long)char_pos, str_len, TRUE);
    if (set_cursor) TXT_SetCursor(atbuf);
    if (char_pos <= atbuf->hdr_end) EDT_RecalcHeaders(atbuf);
  } else {
    SCR_ClearKeyboard();
    SCR_Beep();
  }
}

/*
 * The following routines perform cut / copy / paste data management
 */

char *g_CutBuffer = NIL;

int   g_CutBufferSize = 0;
int   g_CutBufferLen  = 0;

#define CUT_BUFFER_INCREMENTS 1023

#define MAX_COPYLEN 31000

int EDT_ResizeCutBuffer(int len)
{
  if (len >= (g_CutBufferSize - g_CutBufferLen)) {
    char *new_buffer;
    int newlen = g_CutBufferSize + imax(len, CUT_BUFFER_INCREMENTS) + 1;
    if (newlen >= MAX_COPYLEN) return(-1);
    if (g_CutBuffer)
      new_buffer = (char *)realloc(g_CutBuffer, newlen);
    else
      new_buffer = (char *)malloc(newlen);
    if (new_buffer == NIL) return(-1);
    g_CutBuffer = new_buffer;
    g_CutBufferSize = newlen;
  }
  return(0);
}

int EDT_CopyText(char *aText,int len, int append_it)
{
  int rc = 0;
  if (len > 0) {
    if (!append_it) g_CutBufferLen = 0;
    if ((rc = EDT_ResizeCutBuffer(len)) == 0) {
      bcopy(aText, &g_CutBuffer[g_CutBufferLen], len);
      g_CutBufferLen += len;
    }
  }
  return(rc);
}

void EDT_FreeCopyBuffer(void)
{
  if (g_CutBuffer) free(g_CutBuffer);

  g_CutBuffer = NIL;
  g_CutBufferLen = 0;
  g_CutBufferSize = 0;
}

int EDT_GetCopyText(char **text_ptr,int  *text_len)
{
  if (g_CutBuffer) {
    *text_ptr = g_CutBuffer;
    *text_len = g_CutBufferLen;
  } else return(-1);
    
  return(0);
}

int EDT_GetCopyTextLen(void)
{
  if (g_CutBuffer && g_CutBufferLen > 0) return(g_CutBufferLen);
  else return(0);
}

/*
 * header processing routines
 */

int check_for_header(TEXT_BUF_pt atbuf,long pos,int len)
{
  int ret_code = FALSE;

  if (len > 3) {
    char buf[STR255LEN + 1];
    int  bytes = EDT_GetBytes(atbuf, (long)pos, &buf[0], STR255LEN);
    buf[bytes] = 0;
    if (strncmp(&buf[0], "To:", 3) == 0 ||
	strncmp(&buf[0], "Cc:", 3) == 0 ||
	strncmp(&buf[0], "Subject:", 8) == 0) ret_code = TRUE;
  }
  return(ret_code);
}

void EDT_RecalcHeaders(TEXT_BUF_pt atbuf)
{
  Boolean seen_headers = FALSE;
  long pos = 0L, nextpos;
  atbuf->hdr_end = 0L;

  while(pos < (imin(atbuf->fsize, atbuf->bsize))) {
    nextpos = EDT_find_next_eol(atbuf, (long)pos);
    if (check_for_header(atbuf, (long)pos, (int)(nextpos - pos)))
      seen_headers = TRUE;
    if (pos >= nextpos) break;
    pos = nextpos + 1;
  }
  if (seen_headers) atbuf->hdr_end = pos;
  MNU_EnableItem(M_ADD_HEADERS, !seen_headers);
}

void cut_region(TEXT_BUF_pt atbuf,long from, long to)
{
  int len = to - from + 1;
  int bytes = imax(0, atbuf->fsize - from);

  if (len > 0 && bytes > 0) {
    EDT_CloseGap(atbuf);
    bcopy(&atbuf->buf[to], &atbuf->buf[from], bytes);
    atbuf->fsize -= len;
  }
}

void EDT_build_a_header(char *buf,char *hdr_text,int  hdr_index, int need_cr)
{
  strcat(buf, mail_headers[hdr_index]);
  strcat(buf, ": ");
  if (hdr_text) strcat(buf, hdr_text);
  if (need_cr) strcat(buf, "\n");
}

void EDT_AddFaxHeaders(TEXT_BUF_pt atbuf,int use_more_options)
{
  int i;
  char bigbuf[4096];

  if (atbuf->hdr_end > 0) {
    cut_region(atbuf, 0L, (long)atbuf->hdr_end);
  }
  atbuf->char_pos = 0;
  bigbuf[0] = 0;
  EDT_build_a_header(bigbuf, "\"ezfax+", HDR_To, FALSE);
  strcat(bigbuf, g_FaxHeaders->fax_destinfo[FAX_INFO_PHONE]);
  for (i=FAX_INFO_NAME; i<=FAX_INFO_VOICE; i++) {
    if (i == FAX_INFO_PHONE) continue;
    if (g_FaxHeaders->fax_destinfo[i][0]) {
      strcat(bigbuf, "+");
      strcat(bigbuf, g_FaxHeaders->fax_destinfo[i]);
    }
  }

  strcat(bigbuf, "\"\n");

  for (i=4; i<strlen(bigbuf); i++)
    if (bigbuf[i] == SPC_KEY) bigbuf[i] = '_';

  EDT_build_a_header(bigbuf, g_FaxHeaders->fax_destinfo[FAX_INFO_SUBJECT],
		   HDR_Subject, TRUE);
  EDT_build_a_header(bigbuf, NIL, HDR_CC, TRUE);

  if (g_FaxHeaders->fax_destinfo[FAX_INFO_CHARGE][0])
    EDT_build_a_header(bigbuf, g_FaxHeaders->fax_destinfo[FAX_INFO_CHARGE],
		     HDR_Fax_Charge_to, TRUE);

  if (g_FaxHeaders->fax_destinfo[FAX_INFO_TIME][0])
    EDT_build_a_header(bigbuf, g_FaxHeaders->fax_destinfo[FAX_INFO_TIME],
		     HDR_Fax_Delivery_after, TRUE);

  if (use_more_options) {
    EDT_build_a_header(bigbuf,
		     (g_FaxHeaders->fax_bools[FAX_OPT_CONF] ? "yes" : "no"),
		     HDR_Fax_Confidential, TRUE);

    EDT_build_a_header(bigbuf,
		     (g_FaxHeaders->fax_bools[FAX_OPT_RES] ? "fine" : "draft"),
		     HDR_Fax_Resolution, TRUE);

    if (g_FaxHeaders->fax_options[FAX_OPT_RETRY][0])
      EDT_build_a_header(bigbuf, g_FaxHeaders->fax_options[FAX_OPT_RETRY],
		       HDR_Fax_Max_retries, TRUE);

    EDT_build_a_header(bigbuf,
		     (g_FaxHeaders->fax_bools[FAX_OPT_TRAIL] ? "yes" : "no"),
		     HDR_Fax_Trailing_Coverpage_Too, TRUE);

    if (g_FaxHeaders->fax_options[FAX_OPT_BODY][0])
      EDT_build_a_header(bigbuf, g_FaxHeaders->fax_options[FAX_OPT_BODY],
		       HDR_Fax_PostScript_bodyfile, TRUE);

    if (g_FaxHeaders->fax_options[FAX_OPT_COVER][0])
      EDT_build_a_header(bigbuf, g_FaxHeaders->fax_options[FAX_OPT_COVER],
		       HDR_Fax_CoverPage, TRUE);

    if (g_FaxHeaders->fax_bools[FAX_OPT_ACK])
      EDT_build_a_header(bigbuf, do_getuserid(NULL), HDR_Ack_to, TRUE);
  }

  strcat(bigbuf, "\n");

  EDT_InsertString(atbuf, bigbuf, strlen(bigbuf), FALSE);
  atbuf->char_pos = 0;
  EDT_EndCursor(atbuf);
}

void EDT_AddHeader(TEXT_BUF_pt atbuf,int hdr_index,char *hdr_text)
{
  char head_buf[STR255LEN * 2];

  head_buf[0] = 0;

  atbuf->char_pos = 0;
  EDT_build_a_header(head_buf, hdr_text, hdr_index, TRUE);
  EDT_InsertString(atbuf, head_buf,strlen(head_buf),FALSE);
  atbuf->char_pos = 0;
  EDT_HomeCursor(atbuf);
}

long find_start_header(atbuf)
TEXT_BUF_pt atbuf;
{
  long pos = atbuf->char_pos;
  char *colon_pos;
  
  EDT_CloseGap(atbuf);
  colon_pos = index(&atbuf->buf[pos], ':');
  if (colon_pos) {
     char *next_line = index(&atbuf->buf[pos], '\n');
     if (next_line && next_line > colon_pos) {
       pos = colon_pos - &atbuf->buf[0] + 1;
       if (atbuf->buf[pos] == SPC_KEY || atbuf->buf[pos] == TAB_KEY)
         pos++;
     }
  }
  return((long)pos);
}

void EDT_GotoEOL(atbuf)
TEXT_BUF_pt atbuf;
{
  SCR_LINES_pt cur_line = TXT_GetScrLine(atbuf, (long)atbuf->char_pos);

  if (cur_line && cur_line->len > 0) {
    long pos = cur_line->char_pos + cur_line->len;
    if (pos > 0) {
      if (pos < atbuf->fsize) pos--;
      else if (TF_ISSET(cur_line->flags, EDT_F_CRLF)) pos--;
    }
    atbuf->char_pos = pos;
    recalc_mark(atbuf);
  }
}

void EDT_GotoBOL(atbuf)
TEXT_BUF_pt atbuf;
{
  SCR_LINES_pt cur_line = TXT_GetScrLine(atbuf, (long)atbuf->char_pos);

  if (cur_line && cur_line->len > 0) {
    atbuf->char_pos = cur_line->char_pos;
    if (atbuf->char_pos < atbuf->hdr_end) 
      atbuf->char_pos = find_start_header(atbuf);
    else
      recalc_mark(atbuf);
  }
}

void EDT_SaveDraft(atbuf)
TEXT_BUF_pt atbuf;
{
  if (DLG_GetBooleanFromUser("Save message as draft?",TRUE)) { 
    do_draft ( do_getcurrentg(), 
	       config_getstring ( "DraftFolder", 
				  "inbox.drafts" ),
	       atbuf );
    
    /* Mark as unchanged and close the window */
    TF_CLEAR(atbuf->flags, TF_CHANGED);
    TXT_Close ( atbuf->awp );
  }
}
